@@ -452,9 +452,9 @@ gst_jpeg_dec_decode_direct (GstJpegDec * dec, guchar * base[3],
   guchar **line[3];             /* the jpeg line buffer */
   gint i, j, k;
 
-  line[0] = g_alloca ((r_v * DCTSIZE) * sizeof (guchar *));
-  line[1] = g_alloca ((r_v * DCTSIZE) * sizeof (guchar *));
-  line[2] = g_alloca ((r_v * DCTSIZE) * sizeof (guchar *));
+  line[0] = g_malloc ((r_v * DCTSIZE) * sizeof (guchar *));
+  line[1] = g_malloc ((r_v * DCTSIZE) * sizeof (guchar *));
+  line[2] = g_malloc ((r_v * DCTSIZE) * sizeof (guchar *));
   memset (line[0], 0, (r_v * DCTSIZE) * sizeof (guchar *));
   memset (line[1], 0, (r_v * DCTSIZE) * sizeof (guchar *));
   memset (line[2], 0, (r_v * DCTSIZE) * sizeof (guchar *));
@@ -482,6 +482,10 @@ gst_jpeg_dec_decode_direct (GstJpegDec * dec, guchar * base[3],
     }
     jpeg_read_raw_data (&dec->cinfo, line, r_v * DCTSIZE);
   }
+
+  g_free (line[0]);
+  g_free (line[1]);
+  g_free (line[2]);
 }
 
 static void
@@ -594,8 +598,10 @@ gst_jpegdec_chain (GstPad * pad, GstData * _data)
    * copy over the data into our final picture buffer, otherwise jpeglib might
    * write over the end of a line into the beginning of the next line,
    * resulting in blocky artifacts on the left side of the picture. */
-  if ((I420_Y_ROWSTRIDE (width) % (jpegdec->cinfo.max_h_samp_factor *
-              DCTSIZE)) > 0) {
+  if (width % (jpegdec->cinfo.max_h_samp_factor * DCTSIZE) != 0 ||
+      height % (jpegdec->cinfo.max_v_samp_factor * DCTSIZE) != 0) {
+/*(I420_Y_ROWSTRIDE (width) % (jpegdec->cinfo.max_h_samp_factor *
+              DCTSIZE)) > 0) {*/
     gst_jpeg_dec_decode_indirect (jpegdec, base, last, width, height, r_v);
   } else {
     gst_jpeg_dec_decode_direct (jpegdec, base, last, width, height, r_v);
