@@ -58,6 +58,7 @@ struct _GstDecodeBin
   GstElement *typefind;         /* this holds the typefind object */
 
   gboolean threaded;            /* indicating threaded execution is desired */
+  gboolean input_is_stream;
   GList *dynamics;              /* list of dynamic connections */
 
   GList *factories;             /* factories we can use for selecting elements */
@@ -87,6 +88,7 @@ enum
 {
   ARG_0,
   ARG_THREADED,
+  ARG_IS_STREAM
 };
 
 /* signals */
@@ -188,6 +190,10 @@ gst_decode_bin_class_init (GstDecodeBinClass * klass)
   g_object_class_install_property (gobject_klass, ARG_THREADED,
       g_param_spec_boolean ("threaded", "Threaded", "Use threads",
           FALSE, G_PARAM_READWRITE));
+  g_object_class_install_property (gobject_klass,
+      ARG_IS_STREAM, g_param_spec_boolean ("is-stream", "Is stream",
+          "Whether the input is a stream or not; will change seeking behaviour",
+          FALSE, G_PARAM_READWRITE));
 
   gst_decode_bin_signals[SIGNAL_NEW_DECODED_PAD] =
       g_signal_new ("new-decoded-pad", G_TYPE_FROM_CLASS (klass),
@@ -316,6 +322,7 @@ gst_decode_bin_init (GstDecodeBin * decode_bin)
   }
 
   decode_bin->dynamics = NULL;
+  decode_bin->input_is_stream = FALSE;
 }
 
 static void
@@ -546,6 +553,11 @@ try_to_link_1 (GstDecodeBin * decode_bin, GstPad * pad, GList * factories)
         g_signal_connect (element, "got-redirect",
             G_CALLBACK (got_redirect), decode_bin);
       }
+      if (g_object_class_find_property (G_OBJECT_GET_CLASS (element),
+              "is-stream")) {
+        g_object_set (G_OBJECT (element),
+            "is_stream", decode_bin->input_is_stream, NULL);
+      }
 
       /* make sure we catch unlink signals */
       sig = g_signal_connect (G_OBJECT (GST_PAD_REALIZE (pad)), "unlinked",
@@ -919,6 +931,9 @@ gst_decode_bin_set_property (GObject * object, guint prop_id,
     case ARG_THREADED:
       decode_bin->threaded = g_value_get_boolean (value);
       break;
+    case ARG_IS_STREAM:
+      decode_bin->input_is_stream = g_value_get_boolean (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -939,6 +954,9 @@ gst_decode_bin_get_property (GObject * object, guint prop_id, GValue * value,
     case ARG_THREADED:
       g_value_set_boolean (value, decode_bin->threaded);
       break;
+    case ARG_IS_STREAM:
+      g_value_set_boolean (value, decode_bin->input_is_stream);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
