@@ -81,6 +81,7 @@ enum
   /* signals */
   SIGNAL_CLIENT_ADDED,
   SIGNAL_CLIENT_REMOVED,
+  SIGNAL_BUFFER_PROCESSED,
 
   LAST_SIGNAL
 };
@@ -96,6 +97,8 @@ enum
 #define DEFAULT_RECOVER_POLICY		 GST_RECOVER_POLICY_NONE
 #define DEFAULT_TIMEOUT			 0
 #define DEFAULT_SYNC_METHOD		 GST_SYNC_METHOD_NONE
+#define DEFAULT_BUFFERS_TO_SEND           -1
+#define DEFAULT_BUFFER_PROCESSED_SIGNAL   FALSE
 
 enum
 {
@@ -113,12 +116,15 @@ enum
   ARG_BUFFERS_MAX,
   ARG_BUFFERS_SOFT_MAX,
 
+  ARG_BUFFERS_TO_SEND,
+
   ARG_RECOVER_POLICY,
   ARG_TIMEOUT,
   ARG_SYNC_CLIENTS,             /* deprecated */
   ARG_SYNC_METHOD,
   ARG_BYTES_TO_SERVE,
   ARG_BYTES_SERVED,
+  ARG_BUFFER_PROCESSED_SIGNAL
 };
 
 #define GST_TYPE_RECOVER_POLICY (gst_recover_policy_get_type())
@@ -198,6 +204,8 @@ gst_client_status_get_type (void)
     {GST_CLIENT_STATUS_SLOW, "GST_CLIENT_STATUS_SLOW", "Too slow"},
     {GST_CLIENT_STATUS_ERROR, "GST_CLIENT_STATUS_ERROR", "Error"},
     {GST_CLIENT_STATUS_DUPLICATE, "GST_CLIENT_STATUS_DUPLICATE", "Duplicate"},
+    {GST_CLIENT_STATUS_NULLSTATE, "GST_CLIENT_STATUS_NULLSTATE",
+        "Please reconnect later"},
     {0, NULL, NULL},
   };
 
@@ -224,6 +232,7 @@ static void gst_multifdsink_set_property (GObject * object, guint prop_id,
 static void gst_multifdsink_get_property (GObject * object, guint prop_id,
     GValue * value, GParamSpec * pspec);
 
+static void gst_multifdsink_finalize (GObject * object);
 
 static GstElementClass *parent_class = NULL;
 
@@ -294,6 +303,10 @@ gst_multifdsink_class_init (GstMultiFdSinkClass * klass)
       g_param_spec_int ("buffers-soft-max", "Buffers soft max",
           "Recover client when going over this limit (-1 = no limit)", -1,
           G_MAXINT, DEFAULT_BUFFERS_SOFT_MAX, G_PARAM_READWRITE));
+  g_object_class_install_property (G_OBJECT_CLASS (klass), ARG_BUFFERS_TO_SEND,
+      g_param_spec_int ("buffers-to-send", "Buffers to send",
+          "Maximum numbers of frame to send to a client (-1 = no limit)", -1,
+          G_MAXINT, DEFAULT_BUFFERS_TO_SEND, G_PARAM_READWRITE));
 
 #if NOT_IMPLEMENTED
   g_object_class_install_property (G_OBJECT_CLASS (klass), ARG_UNIT_TYPE,
@@ -349,6 +362,12 @@ gst_multifdsink_class_init (GstMultiFdSinkClass * klass)
       g_param_spec_uint64 ("bytes-served", "Bytes served",
           "Total number of bytes send to all clients", 0, G_MAXUINT64, 0,
           G_PARAM_READABLE));
+  g_object_class_install_property (G_OBJECT_CLASS (klass),
+      ARG_BUFFER_PROCESSED_SIGNAL,
+      g_param_spec_boolean ("buffer-processed-signal",
+          "Buffer processed signal",
+          "Emit a signal every time a buffer is sent to a client",
+          DEFAULT_BUFFER_PROCESSED_SIGNAL, G_PARAM_READWRITE));
 
   gst_multifdsink_signals[SIGNAL_ADD] =
       g_signal_new ("add", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
@@ -378,7 +397,15 @@ gst_multifdsink_class_init (GstMultiFdSinkClass * klass)
           client_removed), NULL, NULL, gst_tcp_marshal_VOID__INT_BOXED,
       G_TYPE_NONE, 2, G_TYPE_INT, GST_TYPE_CLIENT_STATUS);
 
+  gst_multifdsink_signals[SIGNAL_BUFFER_PROCESSED] =
+      g_signal_new ("buffer-processed", G_TYPE_FROM_CLASS (klass),
+      G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstMultiFdSinkClass,
+          buffer_processed), NULL, NULL, gst_marshal_VOID__INT, G_TYPE_NONE, 1,
+      G_TYPE_INT);
+
+
   gstelement_class->change_state = gst_multifdsink_change_state;
+  gobject_class->finalize = gst_multifdsink_finalize;
 
   klass->add = gst_multifdsink_add;
   klass->remove = gst_multifdsink_remove;
@@ -388,6 +415,20 @@ gst_multifdsink_class_init (GstMultiFdSinkClass * klass)
   GST_DEBUG_CATEGORY_INIT (multifdsink_debug, "multifdsink", 0, "FD sink");
 }
 
+static void
+gst_multifdsink_finalize (GObject * object)
+{
+  GstMultiFdSink *this;
+
+  this = GST_MULTIFDSINK (object);
+
+  g_array_free (this->bufqueue, TRUE);
+
+  if (G_OBJECT_CLASS (parent_class)->finalize) {
+    G_OBJECT_CLASS (parent_class)->finalize (object);
+  }
+}
+
 static void
 gst_multifdsink_init (GstMultiFdSink * this)
 {
@@ -401,9 +442,8 @@ gst_multifdsink_init (GstMultiFdSink * this)
   this->protocol = DEFAULT_PROTOCOL;
   this->mode = DEFAULT_MODE;
 
-  CLIENTS_LOCK_INIT (this);
+
   this->clients = NULL;
-  this->fd_hash = g_hash_table_new (g_int_hash, g_int_equal);
 
   this->bufqueue = g_array_new (FALSE, TRUE, sizeof (GstBuffer *));
   this->unit_type = DEFAULT_UNIT_TYPE;
@@ -413,6 +453,8 @@ gst_multifdsink_init (GstMultiFdSink * this)
 
   this->timeout = DEFAULT_TIMEOUT;
   this->sync_method = DEFAULT_SYNC_METHOD;
+  this->buffers_to_send = DEFAULT_BUFFERS_TO_SEND;
+  this->buffer_processed_signal = DEFAULT_BUFFER_PROCESSED_SIGNAL;
 }
 
 void
@@ -437,6 +479,7 @@ gst_multifdsink_add (GstMultiFdSink * sink, int fd)
   client->dropped_buffers = 0;
   client->avg_queue_size = 0;
   client->new_connection = TRUE;
+  client->buffers_remaining = sink->buffers_to_send;
 
   /* update start time */
   g_get_current_time (&now);
@@ -447,6 +490,17 @@ gst_multifdsink_add (GstMultiFdSink * sink, int fd)
 
   CLIENTS_LOCK (sink);
 
+  /* don't accept a client in NULL state */
+  if (GST_STATE (sink) == GST_STATE_NULL) {
+    client->status = GST_CLIENT_STATUS_NULLSTATE;
+    CLIENTS_UNLOCK (sink);
+    GST_WARNING_OBJECT (sink, "[fd %5d] element in NULL state, refusing", fd);
+    g_signal_emit (G_OBJECT (sink),
+        gst_multifdsink_signals[SIGNAL_CLIENT_REMOVED], 0, fd, client->status);
+    g_free (client);
+    return;
+  }
+
   /* check the hash to find a duplicate fd */
   clink = g_hash_table_lookup (sink->fd_hash, &client->fd.fd);
   if (clink != NULL) {
@@ -639,15 +693,22 @@ gst_multifdsink_remove_client_link (GstMultiFdSink * sink, GList * link)
   g_slist_foreach (client->sending, (GFunc) gst_data_unref, NULL);
   g_slist_free (client->sending);
   client->sending = NULL;
+  CLIENTS_UNLOCK (sink);
 
   g_signal_emit (G_OBJECT (sink),
       gst_multifdsink_signals[SIGNAL_CLIENT_REMOVED], 0, fd, client->status);
 
+  CLIENTS_LOCK (sink);
   if (!g_hash_table_remove (sink->fd_hash, &client->fd.fd)) {
     GST_WARNING_OBJECT (sink,
         "[fd %5d] error removing client %p from hash", client->fd.fd, client);
   }
-  sink->clients = g_list_delete_link (sink->clients, link);
+  /* after releasing the lock above, the link could be invalid, more
+   * precisely, the next and prev pointers could point to invalid list
+   * links. One optimisation could be to add a cookie to the linked list
+   * and take a shortcut when it did not change between unlocking and locking
+   * our mutex. For now we just walk the list again. */
+  sink->clients = g_list_remove (sink->clients, client);
 
   if (fclass->removed)
     fclass->removed (sink, client->fd.fd);
@@ -794,11 +855,19 @@ gst_multifdsink_client_queue_buffer (GstMultiFdSink * sink,
     gst_multifdsink_client_queue_data (sink, client, (gchar *) header, len);
   }
 
-  GST_LOG_OBJECT (sink, "[fd %5d] queueing buffer of length %d",
-      client->fd.fd, GST_BUFFER_SIZE (buffer));
+  if ((client->buffers_remaining == -1) || (client->buffers_remaining > 0)) {
+    GST_LOG_OBJECT (sink, "[fd %5d] queueing buffer of length %d",
+        client->fd.fd, GST_BUFFER_SIZE (buffer));
 
-  gst_buffer_ref (buffer);
-  client->sending = g_slist_append (client->sending, buffer);
+    gst_buffer_ref (buffer);
+    client->sending = g_slist_append (client->sending, buffer);
+    if (client->buffers_remaining > 0)
+      client->buffers_remaining--;
+  } else {
+    GST_LOG_OBJECT (sink,
+        "[fd %5d] not queueing buffer, client has reached max to send limit",
+        client->fd.fd);
+  }
 
   return TRUE;
 }
@@ -1059,6 +1128,11 @@ gst_multifdsink_handle_client_write (GstMultiFdSink * sink,
           gst_buffer_unref (head);
           /* make sure we start from byte 0 for the next buffer */
           client->bufoffset = 0;
+
+          if (sink->buffer_processed_signal)
+            g_signal_emit (G_OBJECT (sink),
+                gst_multifdsink_signals[SIGNAL_BUFFER_PROCESSED], 0, fd);
+
         }
         /* update stats */
         client->bytes_sent += wrote;
@@ -1509,6 +1583,9 @@ gst_multifdsink_set_property (GObject * object, guint prop_id,
     case ARG_BUFFERS_SOFT_MAX:
       multifdsink->units_soft_max = g_value_get_int (value);
       break;
+    case ARG_BUFFERS_TO_SEND:
+      multifdsink->buffers_to_send = g_value_get_int (value);
+      break;
     case ARG_UNIT_TYPE:
       multifdsink->unit_type = g_value_get_enum (value);
       break;
@@ -1534,6 +1611,9 @@ gst_multifdsink_set_property (GObject * object, guint prop_id,
     case ARG_SYNC_METHOD:
       multifdsink->sync_method = g_value_get_enum (value);
       break;
+    case ARG_BUFFER_PROCESSED_SIGNAL:
+      multifdsink->buffer_processed_signal = g_value_get_boolean (value);
+      break;
 
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -1563,6 +1643,9 @@ gst_multifdsink_get_property (GObject * object, guint prop_id, GValue * value,
     case ARG_BUFFERS_SOFT_MAX:
       g_value_set_int (value, multifdsink->units_soft_max);
       break;
+    case ARG_BUFFERS_TO_SEND:
+      g_value_set_int (value, multifdsink->buffers_to_send);
+      break;
     case ARG_BUFFERS_QUEUED:
       g_value_set_uint (value, multifdsink->buffers_queued);
       break;
@@ -1600,6 +1683,9 @@ gst_multifdsink_get_property (GObject * object, guint prop_id, GValue * value,
     case ARG_BYTES_SERVED:
       g_value_set_uint64 (value, multifdsink->bytes_served);
       break;
+    case ARG_BUFFER_PROCESSED_SIGNAL:
+      g_value_set_boolean (value, multifdsink->buffer_processed_signal);
+      break;
 
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -1618,6 +1704,7 @@ gst_multifdsink_init_send (GstMultiFdSink * this)
   fclass = GST_MULTIFDSINK_GET_CLASS (this);
 
   GST_INFO_OBJECT (this, "starting in mode %d", this->mode);
+  CLIENTS_LOCK_INIT (this);
   this->fdset = gst_fdset_new (this->mode);
 
   if (socketpair (PF_UNIX, SOCK_STREAM, 0, control_socket) < 0) {
@@ -1638,6 +1725,8 @@ gst_multifdsink_init_send (GstMultiFdSink * this)
   this->bytes_to_serve = 0;
   this->bytes_served = 0;
 
+  this->fd_hash = g_hash_table_new (g_int_hash, g_int_equal);
+
   if (fclass->init) {
     fclass->init (this);
   }
@@ -1718,11 +1807,27 @@ gst_multifdsink_change_state (GstElement * element)
     case GST_STATE_PAUSED_TO_READY:
       break;
     case GST_STATE_READY_TO_NULL:
+    {
+      gint i, queuelen;
+
       if (GST_FLAG_IS_SET (sink, GST_MULTIFDSINK_OPEN)) {
         gst_multifdsink_close (GST_MULTIFDSINK (element));
         GST_FLAG_UNSET (sink, GST_MULTIFDSINK_OPEN);
       }
+
+      /* free the queue */
+      queuelen = sink->bufqueue->len;
+      for (i = queuelen - 1; i > 0; i--) {
+        GstBuffer *buf;
+
+        queuelen--;
+        buf = g_array_index (sink->bufqueue, GstBuffer *, i);
+        sink->bufqueue = g_array_remove_index (sink->bufqueue, i);
+
+        gst_buffer_unref (buf);
+      }
       break;
+    }
   }
 
   if (GST_ELEMENT_CLASS (parent_class)->change_state)
